import time
from loguru import logger
from nl2query.db.db_connector import execute_sql_query, execute_snowflake_query


class QueryExecutor:
    def run(self, state):
        """Main function to execute the query executor."""
        start_time = time.time()
        try:
            if state["db_type"] == "postgres":
                query = state["validated_query"]
                result = execute_sql_query(query)
                state["output_response"] = result
            elif state["db_type"] == "snowflake":
                query = state["validated_query"]
                result = execute_snowflake_query(query)
                state["output_response"] = result
            else:
                raise ValueError("Unsupported db_type specified in the state.")

        except Exception as e:
            logger.error(f"Error executing query: {e}")
            state["query_executor_error"] = str(e)
            result = None

        end_time = time.time()
        response_time = end_time - start_time
        state["query_executor_response_time"] = response_time
        logger.info(f"Query executor response time: {response_time:.4f} seconds")

        if result is not None:
            logger.info(f"Output after executing the generated query: {result}")
        else:
            logger.info(
                "Query execution failed. Check 'query_executor_error' for details."
            )

        return state, result



import operator

from typing_extensions import TypedDict
from typing import Annotated, Literal, Dict


class BaseState(TypedDict):
    state_id: int
    module_name: str
    thread_id: str
    version: str
    response_type: Literal["streaming", "regular"]
    errors: Annotated[list, operator.add]
    query: str
    model_type: str
    model_name: str
    temperature: str
    results: str
    user_message: str
    raw_messages: list
    messages: list


class State(BaseState):
    tables_selector_yn: bool
    follow_up_query_yn: bool
    query_reframer_yn: bool
    query_reframer_metadata_yn: bool
    query_reframer_examples_yn: bool
    query_reframer_config_yn: bool
    query_reframer_rag_yn: bool
    refarmed_query: str
    reframed_query_with_metadata: str
    reframed_query_with_config: str
    selected_tables: list
    intent_yn: bool
    rag_yn: bool
    config_mapping: Dict
    query_builder_yn: bool
    query_correcter_yn: bool
    query_executor_yn: bool
    reframed_query: str
    initial_query: str
    validated_query: str
    metadata: Dict
    proceed_to_query_builder_yn: bool
    table_selector_response_time: float
    query_validator_response_time: float
    query_reframer_response_time: float
    query_builder_processing_time: float
    intent_engine_processing_time: float
    query_executor_response_time: float
    query_executor_error: str
    mapping_output: Dict
    selected_table_reason: str
    db_type: Literal["postgres", "snowflake"]
    output_response: str  # output from db



import uuid
from nl2query.agent.graph import NL2QueryGraph
from nl2query.agent.state import State
from nl2query.intent_engine.main import IntentEngine
from nl2query.query_reframer.main import QueryReframer
from nl2query.table_selector.main import TableSelector
from nl2query.query_builder.main import QueryBuilder
from nl2query.query_validator.main import QueryValidator
from nl2query.query_executor.main import QueryExecutor
from nl2query.intent_ambiguity_handler.main import IntentAmbiguityHandler
from dotenv import load_dotenv
from examples.utils import load_json_file, load_txt_file
from loguru import logger


# Define QueryCacher and QueryCacheNode
class QueryCacher:
    def run(self, state):
        logger.info("Query cacher running")
        return state


class QueryCacheNode:
    def __init__(self, query_cache=None):
        self._query_cache = query_cache

    def __call__(self, state, **kwargs):
        try:
            logger.info("I am here at Query cache module: Not implemented yet.")
            state["state_id"] = 11
            self._query_cache.run(state)
            return state
        except Exception as e:
            raise Exception(f"Error at QueryCacheNode: {e}")


query_cache_instance = QueryCacher()


# Define FoxFramework
class FoxFramework(NL2QueryGraph):
    def __init__(self, config_file_path: str = "config.yaml", **kwargs):
        super().__init__(config_file_path=config_file_path, **kwargs)

    def _setup_nodes(self):
        super()._setup_nodes()
        if "query_cache_node" not in self.graph.nodes:
            self.graph.add_node(
                "query_cache_node", QueryCacheNode(query_cache=query_cache_instance)
            )
            logger.info("Added node: query_cache_node")

    def _setup_edges(self):
        super()._setup_edges()
        self.remove_edge("query_reframer_node", "intent_engine_node")
        self.add_conditional_edge(
            "query_reframer_node",
            condition=lambda state: "query_cache_node"
            if state.get("use_cache", True)
            else "query_builder_node",
            targets={"query_cache_node", "query_builder_node", "last_node"},
        )
        self.add_conditional_edge(
            "query_cache_node",
            condition=lambda state: "intent_engine_node"
            if state.get("proceed_to_intent", True)
            else "last_node",
            targets={"intent_engine_node", "last_node"},
        )


class FoxState(State, total=False):
    followup_query: str = None


if __name__ == "__main__":
    load_dotenv()
    # "intent_ambiguity_handler_node",
    interrupt_before = ["query_reframer_node"]
    table_info = load_txt_file("examples/example_fox/input/table_metadata.txt")
    table_selector_prompt = load_txt_file(
        "examples/example_fox/input/table_selector_prompt.txt"
    )
    table_selector_instance = TableSelector(
        prompt=table_selector_prompt,
        table_info=table_info,
        metadata_file_path="examples/example_fox/input/tables_metadata.json",
    )

    # Load prompts and configs
    config_mapping_prompt = load_txt_file(
        "examples/example_fox/input/config_mapping_prompt.txt"
    )
    query_reframer_with_config_prompt = load_txt_file(
        "examples/example_fox/input/query_reframer_with_config_prompt.txt"
    )
    query_reframer_with_metadata_prompt = load_txt_file(
        "examples/example_fox/input/query_reframer_with_metadata_prompt.txt"
    )
    query_reframer_with_config_examples = load_txt_file(
        "examples/example_fox/input/query_reframer_with_config_example.txt"
    )
    config_mapping_example = load_txt_file(
        "examples/example_fox/input/config_mapping_example.txt"
    )
    query_reframer_with_metatdata_examples = load_txt_file(
        "examples/example_fox/input/query_reframer_examples.txt"
    )
    config = load_json_file("examples/example_fox/input/config_mapping_fox.json")
    metadata = load_txt_file("examples/example_fox/input/meta_information_fox.txt")

    query_reframer_instance = QueryReframer(
        config_mapping_prompt=config_mapping_prompt,
        query_reframer_with_config_prompt=query_reframer_with_config_prompt,
        query_reframer_with_metadata_prompt=query_reframer_with_metadata_prompt,
        query_reframer_with_config_examples=query_reframer_with_config_examples,
        config_mapping_example=config_mapping_example,
        query_reframer_with_metatdata_examples=query_reframer_with_metatdata_examples,
        config=config,
        metadata_file_path="examples/example_fox/input/tables_metadata.json",
    )

    intent_prompt = load_txt_file("examples/example_fox/input/intent_engine_prompt.txt")
    intent_examples = load_txt_file(
        "examples/example_fox/input/intent_engine_example.txt"
    )
    intent_engine_instance = IntentEngine(
        system_prompt=intent_prompt,
        examples=intent_examples,
        metadata=metadata,
    )

    query_builder_instance = QueryBuilder(schema_name="GOLD_PERFORMANCE_AND_RATINGS")
    query_validator_instance = QueryValidator()
    query_executor_instance = QueryExecutor()
    intent_ambiguity_instance = IntentAmbiguityHandler(
        ambiguous_fields_file="data/input/ambigious_fields.json"
    )

    # Initialize the graph
    graph = FoxFramework(
        config_file_path="notebooks/config.yaml",
        table_selector_instance=table_selector_instance,
        query_reframer_instance=query_reframer_instance,
        intent_engine_instance=intent_engine_instance,
        query_builder_instance=query_builder_instance,
        query_validator_instance=query_validator_instance,
        query_executor_instance=query_executor_instance,
        intent_ambiguity_instance=intent_ambiguity_instance,
        interrupt_before=interrupt_before,
    )
    state = FoxState()
    # example: What is the average audience rating for cable networks on 2025-03-30?
    query = input("Enter user query:")
    state["query"] = query

    state["model_type"] = "openai"
    state["model_name"] = "gpt-4.1"
    state["temperature"] = "0.1"

    thread_id = str(uuid.uuid4())
    state["thread_id"] = thread_id

    # Initial run
    response = graph.run(
        thread_id=thread_id,
        state=state,
        model_name="openai",
        model_type="gpt-4.1",
        temperature=0.1,
    )

    # snapshot values
    state = response.values
    state_id = response.values.get("state_id") if hasattr(response, "values") else None
    while True:
        current_state_id = state.get("state_id")
        logger.info(f"Current state_id: {current_state_id}")

        if current_state_id in [1, 9, -1]:
            logger.info(" Terminating conversation.")
            print(state.get("selected_tables"))
            break

    # while True:
    #     logger.info(thread_id)
    #     state_id = (
    #         response.values.get("state_id") if hasattr(response, "values") else None
    #     )
    #     if state_id == 4:  # Intent ambiguity
    #         user_message = response.values.get("user_message", "")
    #         user_query = input(f"{user_message} \nPlease enter your clarified query: ")

    #         # Re-run the graph
    #         response = graph.run(
    #             thread_id=thread_id, state=state, updated_state={"query": user_query}
    #         )
    #         state = response.values
    #         print(state)
    #         # break

    #     elif state_id == 9:  # User followup query
    #         followup_yn = input(
    #             "Do you want to ask followup query? If yes, respond with 'y': "
    #         )
    #         if followup_yn.lower() == "y":
    #             followup_query = input("Ask followup query: ")
    #             response = graph.run(
    #                 thread_id=thread_id,
    #                 state=state,
    #                 updated_state={"query": followup_query},
    #             )
    #             logger.info("Graph response after Followup query:", response)
    #         else:
    #             logger.info("User does not want to ask followup query.")
    #             break
    logger.info("Conversation completed.")



from loguru import logger

from nl2query.core.decorators import handle_exceptions


class StartNode:
    def __init__(self, config):
        self._config = config

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            state["state_id"] = 0
            state["module_name"] = "start_node"
            state["conversation_messages"] = []
            state["errors"] = []
            state["query"] = state["query"]
            if not state["query"]:
                raise Exception("Error at StartNode: 'query'")

            state["user_message"] = []
            state["raw_messages"] = []
            state["messages"] = []
            state["config"] = {}
            state["info_message_to_user"] = {}
            state["proceed_to_query_reframer_yn"] = False
            state["metadata_json"] = {}
            state["regenerate_intent_yn"] = False
            state.update(self._config)
            return state
        except Exception as e:
            raise Exception(f"Error at StartNode: {e}")


class TableSelectorNode:
    def __init__(self, table_selector):
        self._table_selector = table_selector

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            state["state_id"] = 1
            state["module_name"] = "table_selector_node"
            _, response = self._table_selector.run(state)
            state["selected_tables"] = response
            return state
        except Exception as e:
            raise Exception(f"Error at TableSelectorNode: {e}")


class QueryReframerNode:
    def __init__(self, query_reframer):
        self._query_reframer = query_reframer

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            state["state_id"] = 2
            state["module_name"] = "query_reframer_node"
            _, response = self._query_reframer.run(state)
            state["reframed_query"] = response
            return state
        except Exception as e:
            raise Exception(f"Error at QueryReframerNode: {e}")


class IntentEngineNode:
    def __init__(self, intent_engine):
        self._intent_engine = intent_engine

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            state["state_id"] = 3
            state["module_name"] = "intent_engine_node"
            _, response = self._intent_engine.run(state)
            state["intent_json"] = response
            return state
        except Exception as e:
            raise Exception(f"Error at IntentEngineNode: {e}")



class QueryBuilderNode:
    def __init__(self, query_builder):
        self._query_builder = query_builder

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            state["state_id"] = 7
            state["module_name"] = "query_builder_node"
            state, response = self._query_builder.run(state)
            state["initial_query"] = response
            return state
        except Exception as e:
            raise Exception(f"Error at QueryBuilderNode: {e}")


class QueryValidatorNode:
    def __init__(self, query_validator):
        self._query_validator = query_validator

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            state["state_id"] = 8
            state["module_name"] = "query_validator_node"
            _, response = self._query_validator.run(state)
            state["validated_query"] = response
            return state
        except Exception as e:
            raise Exception(f"Error at QueryValidatorNode: {e}")


class QueryExecutorNode:
    def __init__(self, query_executor):
        self._query_executor = query_executor

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            state["state_id"] = 9
            state["module_name"] = "query_executor_node"
            _, response = self._query_executor.run(state)
            state["output_response"] = response
            return state
        except Exception as e:
            raise Exception(f"Error at QueryExecutorNode: {e}")


class UserFollowUpNode:
    def __init__(self):
        pass

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            state["state_id"] = 10
            state["module_name"] = "user_followup_query_node"
            state["proceed_to_query_builder_yn"] = False
            state["validated_query"] = None
            state["intent_json"] = None
            state["selected_tables"] = None
            return state
        except Exception as e:
            raise Exception(f"Error at UserFollowUpNode: {e}")


class LastNode:
    def __init__(self):
        pass

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            logger.info("I am here at last node.")
            state["state_id"] = -1
            state["module_name"] = "last_node"
            return state
        except Exception as e:
            raise Exception(f"Error at LastNode: {e}")



from loguru import logger

# from nl2query.agent import State


def wrap_state_transition(func):
    def wrapper(state):
        if state["module_name"] == "last_node":
            return "last_node"
        return func(state)

    return wrapper


def handle_exceptions(func):
    def wrapper(self, state, **kwargs):
        try:
            return func(self, state, **kwargs)
        except Exception as e:
            logger.info(e)
            state["state_id"] = -1
            state["module_name"] == "last_node"
            logger.error(f"An exception occurred: {e}")
            if state.get("errors") is None:
                state["errors"] = []

            # Check if the exception has status_code and detail attributes
            if hasattr(e, "status_code") and hasattr(e, "detail"):
                error_entry = {"status": e.status_code, "detail": str(e.detail)}
            else:
                error_entry = {"status": 500, "detail": str(e)}

            state["errors"].append(error_entry)
            return state

    return wrapper



import msgpack
from langgraph.checkpoint.memory import MemorySaver


class CustomMemorySaver(MemorySaver):
    def serialize(self, state):
        # Remove non-serializable objects before serialization.
        for key in [
            "table_selector_instance",
            "query_reframer_instance",
            "intent_engine_instance",
            "query_builder_instance",
            "query_validator_instance",
            "query_executor_instance",
        ]:
            state.pop(key, None)  # Remove key if it exists.
        try:
            return msgpack.packb(state)
        except Exception as e:
            raise Exception(f"Error during state serialization: {e}")



from pydantic import BaseModel, Field
from typing import Dict, Optional


class IntentJsonSchema(BaseModel):
    """Schema for the intent json"""

    intent_json: Optional[Dict] = Field(
        ...,
        description="A dictionary representing the user's query identified intent that includes the implicitly mentioned Snowflake query fields from the given schema.",
    )



from pydantic import BaseModel, Field
from typing import Dict, Optional


class QueryReframerWithConfigSchema(BaseModel):
    reframed_query: Optional[str] = Field(
        ...,
        description="Reframed natural language query incorporating Snowflake field names or descriptions from given *mapping_json*. Replace original user terms with mapped fields in the generated query, include inline annotations. Ensure that the returned query is never null or an empty string.",
    )


class QueryReframerWithMetadataSchema(BaseModel):
    reframed_query: Optional[str] = Field(
        ...,
        description="Reframed natural language query that integrates Snowflake field names or descriptions derived from the provided *tables metadata*. The query must utilize the field descriptions from the metadata to replace original terms, ensuring alignment with the table's structure.",
    )


class QueryReframerConfigSchema(BaseModel):
    """Schema for the key point."""

    mapping_output: Optional[Dict] = Field(
        ..., description="Config mapping of technical terms to datatbase schema."
    )



from pydantic import BaseModel, Field
from typing import Optional


class QueryValidatorSchema(BaseModel):
    validated_query: Optional[str] = Field(
        None,
        description="A validated Snowflake SQL query string without syntax or logical errors.",
    )



def get_table_selector_prompts(prompt=None, tables_info=None):
    """Function to generate prompts for the pre-filtering engine."""

    if tables_info is None:
        tables_info = ["customers", "products", "order", "order_items"]

    if not prompt:
        prompts = f"""
        Task: Analyze the provided natural language query to determine which tables in the database it references.
        List only the relevant tables from the provided list of tables without giving any explanations or reasoning.

        Input:
        1. Natural Language Query: `query`
        2. **List of Available Tables:**
        {tables_info}

        Output Requirement:
        Provide a list of tables from the provided list that are referenced in the natural language query only if you are sure of it.

        **Guidelines:**
        1. Make sure to list only the relevant tables from the available list.
        2. If no tables are related to the given query, return an empty list.

        Expected Output Format:
        ["ReferencedTable1", "ReferencedTable2", ...]
        (Continue listing all relevant tables identified.)
        """
    else:
        prompts = f"{prompt} \n\n **Tables_info**{tables_info}"

    return str(prompts)



import json
import time

from loguru import logger
from langchain_core.prompts import ChatPromptTemplate

from nl2query.core.base_module import BaseModule
from nl2query.query_builder.schema import QueryBuilderSchema
from nl2query.core.llm_models import get_llm
from nl2query.query_builder.prompts import get_query_builder_prompt


class QueryBuilder(BaseModule):
    """Concrete implementation of BaseModule for intent detection"""

    def __init__(
        self,
        system_prompt: str = None,
        pydantic_class: QueryBuilderSchema = QueryBuilderSchema,
        examples: str = None,
        schema_name: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            system_prompt=system_prompt,
            pydantic_class=pydantic_class,
            examples=examples,
            *args,
            **kwargs,
        )

        self.prompt = system_prompt
        self.examples = examples
        self.pydantic_class = pydantic_class
        self.schema_name = schema_name

        logger.info(f"Initialized IntentEngine with prompt: {system_prompt}")

    def run(self, state):
        """Process the state and return intent JSON"""
        try:
            start_time = time.time()
            model_type = state.get("model_type", "openai")
            model_name = state.get("model_name", "gpt-4-1106-preview")
            temperature = state.get("temperature", 0.01)
            if state["query_reframer_yn"]:
                query = state["reframed_query"]
            else:
                query = state["query"]

            if state["intent_yn"]:
                intent_json = state["intent_json"]

            # TODO: Add both examples with 'intent_json' and 'reframed_query'
            # You can modify this as needed to add examples from your use case.

            prompt = get_query_builder_prompt(
                prompt=self.prompt, schema_name=self.schema_name
            )

            prompt_template = ChatPromptTemplate.from_messages(
                [
                    ("system", f"{prompt}"),
                    ("human", "{query}"),
                    ("system", "{intent_json}"),
                ]
            )
            llm = get_llm(
                model_type=model_type,
                model_name=model_name,
                temperature=temperature,
            )

            structured_llm = llm.with_structured_output(self.pydantic_class)
            few_shot_structured_llm = prompt_template | structured_llm

            query = query.replace("{", "{{").replace("}", "}}")

            response = few_shot_structured_llm.invoke(
                {"query": query, "intent_json": json.dumps(intent_json)}
            )
            response = response.dict()["output_query"]

            initial_query = response
            state["initial_query"] = initial_query
            state["raw_messages"].append({"role": "initial_query", "content": response})

            logger.info(f"Initial query from query builder module: {initial_query}")
            end_time = time.time()
            response_time = end_time - start_time
            state["query_builder_processing_time"] = response_time

            logger.info(f"Query Builder processing time: {response_time:.4f} seconds")

            return state, response

        except Exception as e:
            logger.error(f"Error processing intent: {e}")
            raise



INTENT_ENGINE_PROMPT = (
    "You are an expert at converting natural language queries into structured intent JSON. "
    "Your task is to analyze the provided natural language query and output a structured `intent_JSON` "
    "that identifies the intent and key entities involved."
)

INTENT_STRUCTURE = """
        The structure of the expected `intent_JSON` should be:
            - "intent": (String) The primary action or purpose of the query (e.g., "search", "aggregate", "filter").
            - "entities": (Array of Object) Key entities mentioned in the query with their fields:
            - "field": (String) The schema field referenced.
            - "value": (String, optional) Specific value if mentioned.
            - "operation": (String) The type of operation (e.g., "select", "count", "sum").

            Example:
            Query: "What’s the total sales for 2023?"
            intent_JSON: {{
            "intent": "aggregate",
            "entities": [
                {{"field": "total", "value": null}},
                {{"field": "date", "value": "2023"}}
            ],
            "operation": "sum"
        }}
"""

GUIDELINES = """
Guidelines:
- Identify the main intent of the query (e.g., search, count, aggregate).
- Map any mentioned fields to the schema; if a field isn’t in the schema, include an error in the JSON.
- Return only the `intent_JSON` in valid JSON format, no additional text.
"""


def get_intent_prompt(
    prompt=None,
    examples=None,
    metadata=None,
    intent_structure=None,
    guidelines=None,
    *args,
    **kwargs,
):
    """Generate a simple prompt for intent detection based on the state"""

    intent_engine_prompt = [prompt] if prompt else [INTENT_ENGINE_PROMPT]

    intent_engine_prompt.append(f"**Schema Details**:\n{metadata}")
    intent_engine_prompt.append(
        intent_structure if intent_structure else INTENT_STRUCTURE
    )

    if examples:
        intent_engine_prompt.append(examples)

    intent_engine_prompt.append(guidelines if guidelines else GUIDELINES)
    return "\n\n".join(intent_engine_prompt)



QUERY_VALIDATOR_PROMPT = """
You are an expert SQL query validator. Your task is to double-check the given SQL query for given natural language query and for common mistakes and ensure it is syntactically correct, semantically valid, and safe to execute. Review the query step-by-step and identify any issues, providing a corrected version if applicable.

**Instructions:**
Double-check the given SQL query for the following common mistakes:
1. **Syntax Errors**: Missing or misplaced keywords (e.g., SELECT, FROM, WHERE), unbalanced parentheses, missing semicolons, or incorrect clause order.
2. **Table/Column References**: Ensure all referenced tables and columns exist in the provided schema (if available). Flag undefined tables or columns.
3. **Data Types**: Check for mismatched data types (e.g., comparing strings to integers without proper casting).
4. **Joins**: Verify join conditions are complete (e.g., ON clause present) and logical.

Output the final SQL query only.
"""


def get_query_validator_prompt(prompt=None):
    """Generate a prompt for building an SQL query based on the state."""
    query_validator_prompt = []

    base_prompt = prompt if prompt else QUERY_VALIDATOR_PROMPT
    query_validator_prompt.append(base_prompt)
    return "\n\n".join(query_validator_prompt)



import time
import json
from loguru import logger
from langchain_core.prompts import ChatPromptTemplate

from nl2query.core.base_module import BaseModule
from nl2query.query_reframer.schema import (
    QueryReframerWithConfigSchema,
    QueryReframerConfigSchema,
    QueryReframerWithMetadataSchema,
)
from nl2query.core.llm_models import get_llm
from nl2query.query_reframer.prompts import (
    get_reframed_query_with_config_prompt,
    get_reframed_query_with_metadata_prompt,
    get_config_mapping_prompt,
)


class QueryReframer(BaseModule):
    """Concrete implementation of BaseModule for query reframing with conditional config, metadata, and examples"""

    def __init__(
        self,
        config_mapping_prompt: str = None,
        query_reframer_with_config_prompt: str = None,
        query_reframer_with_metadata_prompt: str = None,
        pydantic_class: QueryReframerWithConfigSchema = QueryReframerWithConfigSchema,
        query_reframer_with_config_examples: str = None,
        config_mapping_example: str = None,
        query_reframer_with_metatdata_examples: str = None,
        config: str = None,
        metadata_file_path: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            pydantic_class=pydantic_class,
            examples=query_reframer_with_config_examples,
            *args,
            **kwargs,
        )
        self.metadata_file_path = metadata_file_path
        self.query = ""
        self.rephrased_query = ""
        self.pydantic_class = pydantic_class
        self.config_mapping_prompt = config_mapping_prompt
        self.query_reframer_with_config_prompt = query_reframer_with_config_prompt
        self.query_reframer_with_metadata_prompt = query_reframer_with_metadata_prompt
        self.query_reframer_with_metadata_examples = (
            query_reframer_with_metatdata_examples
        )
        self.query_reframer_with_config_examples = query_reframer_with_config_examples
        self.config_mapping_example = config_mapping_example
        self.config = config

    def get_previous_msg(self, state):
        previous_messages = ""

        if len(state["messages"]) > 2:
            messages_to_process = (
                state["messages"][-10:]
                if len(state["messages"]) > 10
                else state["messages"]
            )
            for message in messages_to_process:
                role = message["role"]
                content = message["content"]
                if role == "user":
                    previous_messages += f"user: {content}\n"
                elif role == "ai":
                    previous_messages += f"ai: {content}\n"

        previous_messages = str(previous_messages)
        previous_messages = previous_messages.replace("{", "{{").replace("}", "}}")

        return previous_messages

    def load_metadata(self):
        """Load the metadata from the given JSON file path."""
        try:
            with open(self.metadata_file_path, "r") as file:
                metadata = json.load(file)
            return metadata
        except Exception as e:
            logger.error(
                f"Error loading metadata from file {self.metadata_file_path}: {e}"
            )
            raise

    def convert_metadata(self, metadata):
        lines = []
        for table_name, fields in metadata.items():
            lines.append(f"Table: {table_name}")

            table_desc = fields.get("description")
            if table_desc:
                lines.append(f"  Description: {table_desc}\n")

            for field_name, field_info in fields.items():
                if field_name == "description":
                    continue
                data_type = field_info.get("type", "")
                desc = field_info.get("description", "")
                lines.append(f"  {field_name}: {data_type} => {desc}")
            lines.append("")
        return "\n".join(lines)

    def run(self, state):
        """Core logic to reframe the query into an SQL query based on config, metadata, and examples"""
        start_time = time.time()
        # self.metadata = self.load_metadata()
        self.metadata = self.convert_metadata(state["metadata"])
        state["metadata"] = self.metadata
        self.model_type = state.get("model_type", "openai")
        self.model_name = state.get("model_name", "gpt-4-1106-preview")
        self.temperature = state.get("temperature", 0.01)
        self.query = state["query"]

        self.previous_messages = self.get_previous_msg(state)

        if state["query_reframer_metadata_yn"]:
            self.query_reframer_with_metadata_prompt = (
                self.query_reframer_with_metadata_prompt
            )
            self.query_reframer_with_metadata_examples = (
                self.query_reframer_with_metadata_examples
            )
        elif state["query_reframer_config_yn"]:
            self.query_reframer_with_config_prompt = (
                self.query_reframer_with_config_prompt
            )
            self.query_reframer_with_config_examples = (
                self.query_reframer_with_config_examples
            )
        else:
            self.query_reframer_examples = self.examples

        if state["query_reframer_config_yn"]:
            # Get config mapping prompt
            prompt_text = get_config_mapping_prompt(
                self.config,
                self.config_mapping_prompt,
                self.config_mapping_example,
            )

            if isinstance(prompt_text, dict):
                prompt_text = str(prompt_text.get("prompt", ""))

            prompt_template = ChatPromptTemplate.from_messages(
                [("system", prompt_text), ("human", "{query}")]
            )

            llm = get_llm(
                model_type=self.model_type,
                model_name=self.model_name,
                temperature=self.temperature,
            )
            structured_llm = llm.with_structured_output(QueryReframerConfigSchema)

            few_shot_structured_llm = prompt_template | structured_llm

            query = (
                f"Previous messages: {self.previous_messages} \n\n {self.query}"
                if self.previous_messages
                else self.query
            )
            logger.info(f"Input for config mapping: {query}")

            response = few_shot_structured_llm.invoke({"query": query})
            result = response.dict()
            mapping_output = result["mapping_output"]
            state["mapping_output"] = mapping_output
            state["raw_messages"].append(
                {"role": "mapping_output", "content": mapping_output}
            )

            logger.info(f"Mapping Config: {mapping_output}")

            if not mapping_output or mapping_output == {}:
                logger.info("Mapping output is empty, skipping reframing with config.")
            else:
                prompt = get_reframed_query_with_config_prompt(
                    # mapping_output,
                    self.query_reframer_with_config_prompt,
                    self.query_reframer_with_config_examples,
                )
                prompt_template = ChatPromptTemplate.from_messages(
                    [
                        ("system", prompt_text),
                        ("human", "query: {query} \n\n mapping_json: {mapping_json}"),
                    ]
                )
                llm = get_llm(
                    model_type=self.model_type,
                    model_name=self.model_name,
                    temperature=self.temperature,
                )
                structured_llm = llm.with_structured_output(
                    QueryReframerWithConfigSchema
                )

                few_shot_structured_llm = prompt_template | structured_llm
                input_data = {
                    "query": self.query,
                    "mapping_json": json.dumps(mapping_output),
                }
                response = few_shot_structured_llm.invoke(input=input_data)
                # response = few_shot_structured_llm.invoke({"query": self.query})
                reframed_query = response.dict()["reframed_query"]

                logger.info(f"Reframed query with config: {reframed_query}")
                state["reframed_query_with_config"] = reframed_query
                state["reframed_query"] = reframed_query

        if state["query_reframer_metadata_yn"]:
            if not mapping_output or mapping_output == {}:
                logger.info("Mapping output is empty, skipping reframing with config.")
                state["reframed_query_with_config"] = None
                state["reframed_query"] = None
                query = state["query"]

            else:
                query = (
                    state["reframed_query_with_config"]
                    if state.get("query_reframer_config_yn")
                    and state.get("reframed_query_with_config") is not None
                    else state["query"]
                )

            prompt = get_reframed_query_with_metadata_prompt(
                query,
                self.metadata,
                self.query_reframer_with_metadata_prompt,
                self.query_reframer_with_metadata_examples,
            )

            if isinstance(prompt, dict):
                prompt = str(prompt.get("prompt", ""))

            query = (
                f"Previous messages: {self.previous_messages} \n\n {self.query}"
                if self.previous_messages
                else query
            )
            logger.info(f"Input for query reframer: {query}")

            prompt = ChatPromptTemplate.from_messages(
                [("system", prompt), ("human", "{query}")]
            )

            llm = get_llm(
                model_type=self.model_type,
                model_name=self.model_name,
                temperature=self.temperature,
            )
            structured_llm = llm.with_structured_output(
                QueryReframerWithMetadataSchema,
            )
            few_shot_structured_llm = prompt | structured_llm

            response = few_shot_structured_llm.invoke({"query": query})
            rephrased_query = response.dict()["reframed_query"]
            state["reframed_query_with_metadata"] = rephrased_query
            state["reframed_query"] = rephrased_query

            logger.info(f"Reframed query with metadata: {rephrased_query}")

            state["raw_messages"].append(
                {"role": "rephrased_query", "content": self.rephrased_query}
            )
            end_time = time.time()
            response_time = end_time - start_time
            state["query_reframer_response_time"] = response_time
            logger.info(f"Query reframer processing time: {response_time}")

        return state, state["reframed_query"]



from pydantic import BaseModel, Field
from typing import List, Optional


class TableSelectorSchema(BaseModel):
    tables: Optional[List[str]] = Field(
        ..., description="List of tables related to given user query."
    )
    selected_reason: Optional[str] = Field(
        ...,
        description="Justification for the selected tables based on the user query.",
    )



from abc import ABC, abstractmethod

from pydantic import BaseModel
from typing import Optional


class BaseModule(ABC):
    """Abstract base class for intent detection engines that process state and output intent JSON"""

    def __init__(
        self,
        pydantic_class: Optional[BaseModel] = None,
        system_prompt: str = None,
        examples: str = None,
        *args,
        **kwargs
    ):
        self.system_prompt = system_prompt
        self.pydantic_class = pydantic_class
        self.examples = examples or []

    @abstractmethod
    def run(self, state):
        pass



import time

from loguru import logger
from langchain_core.prompts import ChatPromptTemplate

from nl2query.core.base_module import BaseModule
from nl2query.query_validator.schema import QueryValidatorSchema
from nl2query.core.llm_models import get_llm
from nl2query.query_validator.prompts import get_query_validator_prompt


class QueryValidator(BaseModule):
    """Concrete implementation of BaseModule for intent detection"""

    def __init__(
        self,
        system_prompt: str = None,
        pydantic_class: QueryValidatorSchema = QueryValidatorSchema,
        prompt: str = None,
        examples: str = None,
        schema_mapping_path: str = "data/input/schema_mapping.json",
        *args,
        **kwargs,
    ):
        super().__init__(
            system_prompt=system_prompt,
            pydantic_class=pydantic_class,
            examples=examples,
            *args,
            **kwargs,
        )
        self.prompt = prompt
        self.examples = examples
        self.pydantic_class = pydantic_class
        self.schema_mapping_path = schema_mapping_path

        logger.info(f"Initialized IntentEngine with prompt: {system_prompt}")

    def run(self, state):
        """Process the state and return intent JSON"""
        try:
            start_time = time.time()
            model_type = state.get("model_type", "openai")
            model_name = state.get("model_name", "gpt-4o")
            temperature = state.get("temperature", 0.01)

            prompt = get_query_validator_prompt(self.prompt)

            prompt = ChatPromptTemplate.from_messages(
                [("system", prompt), ("human", "{initial_query}")]
            )

            llm = get_llm(
                model_type=model_type,
                model_name=model_name,
                temperature=temperature,
            )
            structured_llm = llm.with_structured_output(self.pydantic_class)
            few_shot_structured_llm = prompt | structured_llm
            initial_query = state["initial_query"]
            response = few_shot_structured_llm.invoke({"initial_query": initial_query})
            validated_query = response.dict()["validated_query"]

            state["validated_query"] = validated_query

            logger.info(f"Query after validation: {validated_query}")
            state["raw_messages"].append(
                {"role": "validated_query", "content": response}
            )

            system_message = {"role": "ai", "content": validated_query}
            state["messages"].append(system_message)

            end_time = time.time()
            response_time = end_time - start_time
            state["query_validator_response_time"] = response_time
            logger.info(f"Query validator processing time: {response_time}")

            return state, validated_query

        except Exception as e:
            logger.error(f"Error processing intent: {e}")
            raise



from pydantic import BaseModel, Field
from typing import Optional


class QueryBuilderSchema(BaseModel):
    output_query: Optional[str] = Field(
        ..., description="Valid SQL query based on given reframed query or intent json"
    )



from contextlib import contextmanager
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy import text
import snowflake.connector

from nl2query import settings


@contextmanager
def get_db():
    """Context manager for getting a database session."""
    # Set up the database engine and session
    engine = create_engine(settings.DATABASE_URL)
    SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)

    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()


def execute_sql_query(query):
    """Executes a raw SQL query on the PostgreSQL database using a session from get_db."""
    try:
        # Use the context manager to manage the session
        with get_db() as db:
            result = db.execute(text(query))  # Execute query using session
            if query.strip().upper().startswith("SELECT"):
                return result.fetchall()
            else:
                print("Query executed successfully!")
                return None
    except SQLAlchemyError as e:
        print(f"Error executing query: {e}")


def create_snowflake_connection():
    conn = snowflake.connector.connect(
        user=settings.SNOWFLAKE_USER,
        password=settings.SNOWFLAKE_PASSWORD,
        account=settings.SNOWFLAKE_ACCOUNT,
        warehouse=settings.SNOWFLAKE_WAREHOUSE,
        database=settings.SNOWFLAKE_DATABASE,
        schema=settings.SNOWFLAKE_SCHEMA,
    )
    return conn


def execute_snowflake_query(query):
    conn = create_snowflake_connection()
    try:
        cur = conn.cursor()
        cur.execute(query)
        results = cur.fetchall()

        return results
    except Exception as e:
        print("Error executing query:", e)
        raise e

    finally:
        cur.close()
        conn.close()
